import toml
from schema import Schema, Optional, And, Or, Use


def positive(number):
    assert number > 0
    return number


schema = Schema(
    {
        "order": [str],
        Optional("click_events"): bool,
        Optional("include"): [str],
        Optional("interval"): Use(float),
        Optional("settings"): Or(
            {
                str: Or(
                    {
                        Optional(str): object,
                    },
                    {},
                ),
            },
            {},
        ),
        Optional("on_click"): Or(
            {
                str: Or(
                    {
                        And(Use(int), Use(positive)): Or([Use(str)], Use(str)),
                    },
                    {},
                )
            },
            {},
        ),
    },
    ignore_extra_keys=True,
)


class Config(dict):
    def read_file(self, file):
        self.update(schema.validate(toml.loads(open(file).read())))



"""Generates a status line for swaybar"""

import os, sys
from logging import Formatter, StreamHandler, FileHandler
from pathlib import Path
from argparse import ArgumentParser
from .loop import run
from .config import Config
from .modules import Modules
from .logging import logger

me = os.path.basename(sys.argv[0])
log_formatter = Formatter("%(levelname)s: %(message)s")
config_home = Path(os.environ.get("XDG_CONFIG_HOME", "~/.config")).expanduser()

try:
    from systemd.journal import JournalHandler
except ModuleNotFoundError:
    journal_available = False
else:
    journal_available = True


def environ_path(name, default=None):
    value = os.environ.get(name, default)
    return Path(value).expanduser() if value else default


def environ_paths(name):
    value = os.environ.get(name)
    parts = value.split(":") if value else []
    return [Path(p).expanduser() for p in parts]


def add_log_handler(handler):
    handler.setFormatter(log_formatter)
    logger.addHandler(handler)


def parse_args():
    p = ArgumentParser(description=__doc__)

    p.add_argument(
        "-c",
        "--config-file",
        metavar="FILE",
        type=Path,
        help="override configuration file",
    )

    p.add_argument(
        "-C",
        "--config-dir",
        metavar="DIRECTORY",
        type=Path,
        help="override configuration directory",
    )

    p.add_argument(
        "-I",
        "--include",
        action="append",
        metavar="DIRECTORY",
        type=Path,
        help="include additional module package",
    )

    p.add_argument(
        "-i",
        "--interval",
        type=float,
        metavar="SECONDS",
        help="specify interval between updates",
    )

    p.add_argument(
        "--no-click-events",
        dest="click_events",
        action="store_false",
        help="disable click events",
    )

    p.add_argument(
        "-l",
        "--log-file",
        metavar="FILE",
        help="output logging to %(metavar)s (default: stderr)",
    )

    if journal_available:
        p.add_argument(
            "-j",
            "--log-journal",
            action="store_true",
            help="output logging to systemd journal",
        )

    return p.parse_args()


def main():
    args = parse_args()

    config_dir = args.config_dir or environ_path(
        "SWAYSTATUS_CONFIG_DIR", config_home / me
    )
    config_file = args.config_file or environ_path(
        "SWAYSTATUS_CONFIG_FILE", config_dir / "config.toml"
    )

    config = Config()
    config.read_file(config_file)

    add_log_handler(StreamHandler())

    if args.log_file:
        add_log_handler(FileHandler(args.log_file))

    if args.log_journal:
        add_log_handler(JournalHandler(SYSLOG_IDENTIFIER=logger.name))

    config["include"] = (
        (args.include or [])
        + [config_dir / "modules"]
        + [Path(d).expanduser() for d in config.get("include", [])]
        + environ_paths("SWAYSTATUS_MODULE_PATH")
    )

    if args.interval:
        config["interval"] = args.interval

    if not args.click_events:
        config["click_events"] = False

    elements = []
    modules = Modules(config["include"])
    settings = config.get("settings", {})

    for module_id in config.get("order", []):
        module_name = module_id.split(":", maxsplit=1)[0]
        element_settings = settings.get(module_name, {}).copy()
        element_settings.update(settings.get(module_id, {}))
        element = modules.find(module_name).Element(**element_settings)
        element.name = module_id
        elements.append(element)

    try:
        run(elements, **config)
    except Exception as e:
        logger.exception("unhandled exception in main loop")
        sys.exit(1)



import sys
from uuid import uuid4
from pathlib import Path
from importlib import import_module, metadata
from importlib.util import spec_from_file_location, module_from_spec


class Modules:
    def __init__(self, include):
        self._packages = []
        self._cache = {}

        for i, modules_dir in enumerate(include):
            package_name = str(uuid4()).replace("-", "_")
            package_init_file = Path(modules_dir) / "__init__.py"
            if package_init_file.is_file():
                spec = spec_from_file_location(package_name, package_init_file)
                if spec:
                    package = module_from_spec(spec)
                    sys.modules[package_name] = package
                    spec.loader.exec_module(package)
                    self._packages.append(package_name)

        entry_points = metadata.entry_points().get("swaystatus.modules", [])

        for entry_point in entry_points:
            self._packages.append(entry_point.load().__name__)

    def find(self, name):
        if name not in self._cache:
            for package in self._packages:
                try:
                    self._cache[name] = import_module(f"{package}.{name}")
                    break
                except ModuleNotFoundError:
                    continue
            else:
                raise ModuleNotFoundError(
                    f"No module named '{name}' in any modules package"
                )

        return self._cache[name]



import sys, locale, json
from signal import signal, SIGUSR1
from threading import Thread
from _thread import interrupt_main
from .updater import Updater
from .logging import logger


def run(elements, **config):
    locale.setlocale(locale.LC_ALL, "")

    elements_by_name = {
        element.name: element
        for element in elements
        if hasattr(element, "name")
    }

    updater = Updater(elements, **config)

    def stdout():
        try:
            updater.run()
        except Exception as e:
            logger.exception("unhandled exception in output thread")
            interrupt_main()

    def update(*args, **kwargs):
        try:
            updater.update()
        except Exception as e:
            logger.exception("unhandled exception when updating")
            sys.exit(1)

    signal(SIGUSR1, update)

    stdout_thread = Thread(target=stdout)
    stdout_thread.daemon = True
    stdout_thread.start()

    # Discard the opening '['.
    sys.stdin.readline()

    for line in sys.stdin:
        click_event = json.loads(line.lstrip(","))
        try:
            elements_by_name[click_event["name"]].on_click(click_event)
        except Exception as e:
            logger.exception("unhandled exception during click event")
            sys.exit(1)



import os
import subprocess
from types import MethodType


class Element:
    name = None

    def __init__(self, *args, **kwargs):
        super().__init__()

        self.intervals = []

        self.env = kwargs.get("env", {})

        for button, handler in kwargs.get("on_click", {}).items():
            self._set_on_click_handler(button, handler)

    def _set_on_click_handler(self, button, handler):
        if not callable(handler):

            def method(self, event):
                env = os.environ.copy()
                env.update(self.env)
                env.update(
                    {
                        key: str(value if value is not None else "")
                        for key, value in event.items()
                    }
                )
                subprocess.run(handler, shell=True, env=env)

        else:

            def method(self, event):
                handler(event)

        setattr(self, f"on_click_{button}", MethodType(method, self))

    def create_block(self, full_text, **params):
        block = {"full_text": full_text}
        block.update(params)

        if self.name:
            block["name"] = self.name

        return block

    def set_interval(self, seconds, options=None):
        self.intervals.append((seconds, options))

    def on_interval(self, options=None):
        pass

    def on_update(self, output):
        pass

    def on_click(self, event):
        try:
            getattr(self, f"on_click_{event['button']}")(event)
            self.updater.update()
        except AttributeError:
            pass



import json
import time
from signal import SIGTSTP


class Updater:
    def __init__(self, elements, **options):
        super().__init__()

        self.elements = elements
        self.element_timers = []

        for element in elements:
            element.updater = self
            self.element_timers.append([0] * len(element.intervals))

        self.interval = options.get("interval", 1)

        self.time_before = time.perf_counter()

        self._header = {
            "version": 1,
            "click_events": options.get("click_events", True),
            "stop_signal": SIGTSTP,
        }
        self._body_start = "[[]"
        self._body_item = ",{}"

    def _running(self):
        return True

    def _send_line(self, line):
        print(line, flush=True)

    def update(self):
        time_now = time.perf_counter()
        self.seconds_elapsed = time_now - self.time_before
        self.time_before = time_now

        output = []

        for element_index, element in enumerate(self.elements):
            timers = self.element_timers[element_index]
            for interval_index, timer in enumerate(timers):
                timer += self.seconds_elapsed
                interval, options = element.intervals[interval_index]
                if timer >= interval:
                    element.on_interval(options=options)
                    timers[interval_index] = 0
                else:
                    timers[interval_index] = timer
            element.on_update(output)

        self._send_line(self._body_item.format(json.dumps(output)))

    def run(self):
        self._send_line(json.dumps(self._header))
        self._send_line(self._body_start)

        while self._running():
            self.update()
            time.sleep(self.interval)
